var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _AbstractGossipLog_apply;
import { TypedEventEmitter, CustomEvent } from "@libp2p/interface";
import { logger } from "@libp2p/logger";
import { ed25519 } from "@canvas-js/signatures";
import { assert } from "@canvas-js/utils";
import { Driver } from "./sync/driver.js";
import { decodeId, encodeId, encodeSignedMessage, getNextClock, messageIdPattern, MIN_MESSAGE_ID, MAX_MESSAGE_ID, decodeSignedMessage, } from "./schema.js";
import { topicPattern } from "./utils.js";
export class AbstractGossipLog extends TypedEventEmitter {
    constructor(init) {
        super();
        _AbstractGossipLog_apply.set(this, void 0);
        assert(topicPattern.test(init.topic), "invalid topic (must match [a-zA-Z0-9\\.\\-])");
        this.topic = init.topic;
        this.indexAncestors = init.indexAncestors ?? false;
        this.signer = init.signer ?? ed25519.create();
        __classPrivateFieldSet(this, _AbstractGossipLog_apply, init.apply, "f");
        this.validatePayload = init.validatePayload ?? ((payload) => true);
        this.verifySignature = init.verifySignature ?? this.signer.scheme.verify;
        this.log = logger(`canvas:gossiplog:[${this.topic}]`);
    }
    async replay() {
        await this.read(async (txn) => {
            for await (const leaf of txn.messages.nodes(0, { key: null, inclusive: false })) {
                assert(leaf.key !== null, "expected leaf.key !== null");
                assert(leaf.value !== undefined, "expected leaf.value !== undefined");
                const [id, signature, message] = this.decode(leaf.value);
                assert(id === decodeId(leaf.key), "expected id === decodeId(key)");
                await __classPrivateFieldGet(this, _AbstractGossipLog_apply, "f").apply(txn, [id, signature, message]);
            }
        });
    }
    async *iterate(lowerBound = null, upperBound = null, options = {}) {
        const { id: lowerId, inclusive: lowerInclusive } = lowerBound ?? { id: MIN_MESSAGE_ID, inclusive: true };
        const { id: upperId, inclusive: upperInclusive } = upperBound ?? { id: MAX_MESSAGE_ID, inclusive: true };
        assert(messageIdPattern.test(lowerId), "lowerBound.id: invalid message ID");
        assert(messageIdPattern.test(upperId), "upperBound.id: invalid message ID");
        for await (const [key, value] of this.entries({ key: encodeId(lowerId), inclusive: lowerInclusive }, { key: encodeId(upperId), inclusive: upperInclusive }, options)) {
            const [id, signature, message] = this.decode(value);
            assert(id === decodeId(key), "expected id === decodeId(key)");
            yield [id, signature, message];
        }
    }
    encode(signature, message) {
        assert(this.topic === message.topic, "expected this.topic === topic");
        assert(this.validatePayload(message.payload), "error encoding message (invalid payload)");
        return encodeSignedMessage(signature, message);
    }
    decode(value) {
        const [id, signature, { topic, clock, parents, payload }] = decodeSignedMessage(value);
        assert(this.topic === topic, "expected this.topic === topic");
        assert(this.validatePayload(payload), "error decoding message (invalid payload)");
        return [id, signature, { topic, clock, parents, payload }];
    }
    async getClock() {
        const heads = await this.read((txn) => txn.getHeads());
        const clock = getNextClock(heads);
        return [clock, heads.map(decodeId)];
    }
    async has(id) {
        assert(messageIdPattern.test(id), "invalid message ID");
        const leaf = await this.read((txn) => txn.messages.getNode(0, encodeId(id)));
        return leaf !== null;
    }
    async get(id) {
        assert(messageIdPattern.test(id), "invalid message ID");
        const leaf = await this.read((txn) => txn.messages.getNode(0, encodeId(id)));
        if (leaf === null) {
            return [null, null];
        }
        assert(leaf.value !== undefined, "expected leaf.value !== undefined");
        const [recoveredId, signature, message] = decodeSignedMessage(leaf.value);
        assert(recoveredId === id, "expected recoveredId === id");
        return [signature, message];
    }
    /**
     * Sign and append a new *unsigned* message to the end of the log.
     * The currently unmerged heads of the local log are used as parents.
     */
    async append(txn, payload, options = {}) {
        const signer = options.signer ?? this.signer;
        const heads = await txn.getHeads();
        const clock = getNextClock(heads);
        const parents = heads.map(decodeId);
        const message = { topic: this.topic, clock, parents, payload };
        const signature = await signer.sign(message);
        const [key, value] = this.encode(signature, message);
        const id = decodeId(key);
        this.log("appending message %s: %O", id, message);
        await this.apply(txn, id, signature, message, [key, value]);
        return { id, signature, message };
    }
    async getMissingParents(txn, parents) {
        const missingParents = new Set();
        this.log("looking up %s parents", parents.length);
        for (const parentId of parents) {
            // TODO: txn.messages.getMany
            const leaf = await txn.messages.getNode(0, encodeId(parentId));
            if (leaf !== null) {
                this.log("found parent %s", parentId);
            }
            else {
                this.log("missing parent %s", parentId);
                missingParents.add(parentId);
            }
        }
        return missingParents;
    }
    /**
     * Insert an existing signed message into the log (ie received via HTTP API).
     * If any of the parents are not present, throw an error.
     */
    async insert(txn, signature, message, [key, value] = this.encode(signature, message)) {
        const id = decodeId(key);
        this.log("inserting message %s", id);
        await this.verifySignature(signature, message);
        // TODO: txn.hasMany(keys)
        for (const parent of message.parents) {
            const leaf = await txn.messages.getNode(0, encodeId(parent));
            if (leaf === null) {
                this.log.error("missing parent %s of message %s: %O", parent, id, message);
                throw new Error(`missing parent ${parent} of message ${id}`);
            }
        }
        await this.apply(txn, id, signature, message, [key, value]);
        return { id };
    }
    async apply(txn, id, signature, message, [key, value] = this.encode(signature, message)) {
        this.log("applying %s %O", id, message);
        try {
            await __classPrivateFieldGet(this, _AbstractGossipLog_apply, "f").apply(txn, [id, signature, message]);
        }
        catch (error) {
            this.dispatchEvent(new CustomEvent("error", { detail: { error } }));
            throw error;
        }
        await txn.insert(signature, message, [key, value]);
        this.dispatchEvent(new CustomEvent("message", { detail: { id, signature, message } }));
    }
    async getAncestors(txn, id, atOrBefore) {
        const results = new Set();
        await txn.getAncestors(encodeId(id), atOrBefore, results);
        this.log("getAncestors of %s atOrBefore %d: %o", id, atOrBefore, results);
        return Array.from(results).sort();
    }
    async isAncestor(txn, id, ancestor) {
        return await txn.isAncestor(encodeId(id), encodeId(ancestor));
    }
    /**
     * Sync with a remote source, applying and inserting all missing messages into the local log
     */
    async *sync(txn, source, options = {}) {
        const start = performance.now();
        let messageCount = 0;
        const driver = new Driver(this.topic, source, txn.messages);
        for await (const [key, value] of driver.sync()) {
            const [id, signature, message] = this.decode(value);
            assert(id === decodeId(key), "expected id === decodeId(key)");
            await this.verifySignature(signature, message);
            // TODO: txn.has(key)
            // TODO: unnecessary????
            const leaf = await txn.messages.getNode(0, key);
            if (leaf !== null) {
                continue;
            }
            yield [id, signature, message, [key, value]];
            messageCount++;
        }
        const duration = Math.ceil(performance.now() - start);
        this.log("finished sync with peer %s (%d messages in %dms)", options.sourceId, messageCount, duration);
        this.dispatchEvent(new CustomEvent("sync", { detail: { peerId: options.sourceId, messageCount, duration } }));
    }
    /**
     * Serve a read-only snapshot of the merkle tree
     */
    async serve(callback, options = {}) {
        await this.read((txn) => callback(txn.messages), options);
    }
}
_AbstractGossipLog_apply = new WeakMap();
